// Conn.java - ACAP connection implementation
// Rob Earhart
// $Id$
//
/***********************************************************
        Copyright 1997 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

package acap;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Vector;

/**
 * A protocol-level interface to the
 * Application Configuration Access Protocol, ACAP.
 * @version 1.0
 * @author Rob Earhart
 * @author Carnegie Mellon University
 */
public class Conn {
  private Parser parser;
  private DataOutputStream out;
  private int tag = 0;
  Hashtable tags;
  Hashtable contexts;

  private Conn(Socket conn) throws java.io.IOException {
    out = new DataOutputStream(new BufferedOutputStream(conn.getOutputStream()));
    tags = new Hashtable();
    contexts = new Hashtable();
    parser = new Parser(conn.getInputStream(), tags, contexts);
    parser.start();
  }

  /**
   * Creates an ACAP Connection to a host
   * @param host the name of the host to connect to
   */
  public Conn(String host) throws java.io.IOException {
    this(new Socket(host, 144));
  }
  /**
   * Creates an ACAP Connection to an address
   * @param address the address to connect to
   * @see java.net.InetAddress
   */
  public Conn(InetAddress address) throws java.io.IOException {
    this(new Socket(address, 144));
  }
  private String newTag() {
    return ("a" + ++tag);
  }
  synchronized void sendCommand(Sendable cmd[])
    throws java.io.IOException {
      try {
	cmd[0].Send(out);
	for(int lupe = 1; lupe < cmd.length; lupe++) {
	  out.writeByte(' ');
	  cmd[lupe].Send(out);
	}
      } finally {
	out.writeBytes("\r\n");
      }
      out.flush();
  }

  /**
   * Sends a login command
   * @param name the name to login as
   * @param password the password to use
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   */
  public String login(String name, String password)
    throws java.io.IOException, acapException {
      String tag = newTag();
      Callback cb = new Callback(null);
      Sendable cmd[] = {new SendTag(tag), new SendAtom("Login"),
			new SendString(name), new SendString(password)};
      return cb.Wait(this, tag, cmd);
  }
  /**
   * Sends a store command
   * @param dataset the dataset to store into
   * @param entry the entry in that dataset
   * @param data the data to be stored
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   */
  public String store(String dataset, String entry, String data[])
    throws java.io.IOException, acapException {
      String tag = newTag();
      Callback cb = new Callback(null);
      Sendable cmd[] = new Sendable[data.length + 4];
      cmd[0] = new SendTag(tag);
      cmd[1] = new SendAtom("Store");
      cmd[2] = new SendString(dataset);
      cmd[3] = new SendString(entry);
      for (int lupe = 0; lupe < data.length; lupe++)
	cmd[lupe + 4] = new SendString(data[lupe]);
      return cb.Wait(this, tag, cmd);
  }

  /**
   * Sends a search command, and places the search results
   * into the supplied vector.
   * @param searchable the dataset/context to search
   * @param modifiers search modifiers
   * @param criteria search criteria
   * @param result the vector to add the results onto
   * @param modtime the place to put the modtime
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   * @see acap.Modifiers
   * @see acap.Criteria
   * @see java.util.Vector
   */
  public String search(String searchable, Modifiers modifiers,
		       Criteria criteria, Vector result, StringBuffer modtime)
  throws java.io.IOException, acapException {
    String tag = newTag();
    Callback cb = new Callback(new VectorSearchHandler(result, modtime));
    Sendable cmd[] = new Sendable[5];
    cmd[0] = new SendTag(tag);
    cmd[1] = new SendAtom("Search");
    cmd[2] = new SendString(searchable);
    cmd[3] = modifiers;
    cmd[4] = criteria;
    return cb.Wait(this, tag, cmd);
  }

  /**
   * Sends a search command, using the specified SearchHandler
   * to deal with intermediate results.
   * @param searchable the dataset/context to search
   * @param modifiers search modifiers
   * @param criteria search criteria
   * @param handler search result handler
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   * @see acap.Modifiers
   * @see acap.Criteria
   * @see acap.SearchHandler
   */
  public String search(String searchable, Modifiers modifiers,
		       Criteria criteria, SearchHandler handler)
  throws java.io.IOException, acapException {
    String tag = newTag();
    Callback cb = new Callback(handler);
    Sendable cmd[] = new Sendable[5];
    cmd[0] = new SendTag(tag);
    cmd[1] = new SendAtom("Search");
    cmd[2] = new SendString(searchable);
    cmd[3] = modifiers;
    cmd[4] = criteria;
    return cb.Wait(this, tag, cmd);
  }

  /**
   * Sends a notifycontext request
   * @param context the context to enable notification on
   * @param attributes the list of attributes to monitor
   * @param handler the object that will deal with notification handling
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   * @see acap.NotifyHandler
   */
  public String notifycontext(String context, String attributes[],
			      NotifyHandler handler)
    throws java.io.IOException, acapException {
      String tag = newTag();
      Callback cb = new Callback(context, handler, true);
      Sendable cmd[] = new Sendable[attributes.length + 3];
      cmd[0] = new SendTag(tag);
      cmd[1] = new SendAtom("NotifyContext");
      cmd[2] = new SendString(context);
      for (int lupe = 0; lupe < attributes.length; lupe++)
	cmd[lupe + 3] = new SendString(attributes[lupe]);
      return cb.Wait(this, tag, cmd);
  }

  /**
   * Sends a freecontext request, disabling notification for
   * the given context.  Note that notification is not disabled
   * until the command completes; notifications sent before
   * the freecontext completes will be handled as normal.
   * @param context the context to free
   * @return the server's operation result string
   * @exception acapException when a problem occurs on the server side
   */
  public String freecontext(String context)
    throws java.io.IOException, acapException {
      String tag = newTag();
      Callback cb = new Callback(context, null, false);
      Sendable cmd[] = new Sendable[3];
      cmd[0] = new SendTag(tag);
      cmd[1] = new SendAtom("FreeContext");
      cmd[2] = new SendString(context);
      return cb.Wait(this, tag, cmd);
  }

  /**
   * Gracefully tears down the ACAP connection when
   * garbage is collected.
   * @exception acapException when a problem occurs on the server side
   */
  protected void finalize()
    throws java.io.IOException, acapException {
      String tag = newTag();
      Callback cb = new Callback(null);
      Sendable cmd[] = new Sendable[2];
      cmd[0] = new SendTag(tag);
      cmd[1] = new SendAtom("Logout");
      cb.Wait(this, tag, cmd);
  }
}

class Parser extends Thread {
  private Yylex lexer;
  private Hashtable tags;
  private Hashtable contexts;
  public Parser(java.io.InputStream in, Hashtable tags, Hashtable contexts) {
    lexer = new Yylex(in);
    this.tags = tags;
    this.contexts = contexts;
    setName("ACAP Parser");
    setDaemon(true);
    setPriority(NORM_PRIORITY - 1);
  }
  public void run() {
    try {
      while(true) {
	try {
	  Yytoken tagtok = lexer.yylex();
	  tagtok.Assert(Yytoken.TAG);
	  if (tagtok == null) {
	    return;
	  }
	  String tag = tagtok.Str();
	  if (tag.equals("*")) {
	    // Untagged response
	    int state = lexer.yylex().Tok();
	    switch(state) {
	    case Yytoken.ADDTO:
	      {
		Yytoken context = lexer.yylex();
		context.Assert(Yytoken.STRING);
		Yytoken entry = lexer.yylex();
		entry.Assert(Yytoken.STRING);
		Yytoken pos = lexer.yylex();
		pos.Assert(Yytoken.INT);
		Vector v = new Vector();
		while(true) {
		  Yytoken tok = lexer.yylex();
		  if (tok.Tok() == Yytoken.EOL)
		    break;
		  tok.Assert(Yytoken.STRING);
		  v.addElement(tok.Str());
		}
		NotifyHandler handler =
		  (NotifyHandler)contexts.get(context.Str());
		if (handler != null) {
		  handler.AddTo(entry.Str(), pos.Int() - 1, v);
		}
		break;
	      }
	    case Yytoken.REMOVEFROM:
	      {
		Yytoken context = lexer.yylex();
		context.Assert(Yytoken.STRING);
		Yytoken entry = lexer.yylex();
		entry.Assert(Yytoken.STRING);
		Yytoken pos = lexer.yylex();
		pos.Assert(Yytoken.INT);
		NotifyHandler handler =
		  (NotifyHandler)contexts.get(context.Str());
		if (handler != null) {
		  handler.RemoveFrom(entry.Str(), pos.Int() - 1);
		}
		lexer.yylex().Assert(Yytoken.EOL);
		break;
	      }
	    case Yytoken.CHANGE:
	      {
		Yytoken context = lexer.yylex();
		context.Assert(Yytoken.STRING);
		Yytoken entry = lexer.yylex();
		entry.Assert(Yytoken.STRING);
		Yytoken oldpos = lexer.yylex();
		oldpos.Assert(Yytoken.INT);
		Yytoken newpos = lexer.yylex();
		newpos.Assert(Yytoken.INT);
		Vector v = new Vector();
		while(true) {
		  Yytoken tok = lexer.yylex();
		  if (tok.Tok() == Yytoken.EOL)
		    break;
		  tok.Assert(Yytoken.STRING);
		  v.addElement(tok.Str());
		}
		NotifyHandler handler =
		  (NotifyHandler)contexts.get(context.Str());
		if (handler != null) {
		  handler.Change(entry.Str(), oldpos.Int() - 1,
				 newpos.Int() - 1, v);
		}
		break;
	      }
	    case Yytoken.MODTIME:
	      {
		Yytoken context = lexer.yylex();
		context.Assert(Yytoken.STRING);
		Yytoken modtime = lexer.yylex();
		modtime.Assert(Yytoken.STRING);
		NotifyHandler handler =
		  (NotifyHandler)contexts.get(context.Str());
		if (handler != null) {
		  handler.ModTime(modtime.Str());
		}
		lexer.yylex().Assert(Yytoken.EOL);
		break;
	      }
	    default:
	      lexer.Resync();
	      break;
	    }
	  } else {
	    // Tagged response
	    int state = lexer.yylex().Tok();
	    switch(state) {
	    case Yytoken.OK:
	    case Yytoken.NO:
	    case Yytoken.BAD:
	    case Yytoken.BYE:
	      {
		String data = lexer.yylex().Str();
		Callback cb = (Callback)tags.get(tag);
		if (cb != null) {
		  cb.Complete(state, data);
		}
		break;
	      }
	    case Yytoken.ENTRY:
	      {
		Vector v = new Vector();
		while(true) {
		  Yytoken tok = lexer.yylex();
		  if (tok.Tok() == Yytoken.EOL)
		    break;
		  tok.Assert(Yytoken.STRING);
		  v.addElement(tok.Str());
		}
		Callback cb = (Callback)tags.get(tag);
		if (cb != null
		    && cb.searchHandler() != null)
		  cb.searchHandler().Entry(v);
		break;
	      }
	    case Yytoken.MODTIME:
	      {
		Yytoken tok = lexer.yylex();
		tok.Assert(Yytoken.STRING);
		Callback cb = (Callback)tags.get(tag);
		if (cb != null
		    && cb.searchHandler() != null)
		  cb.searchHandler().Searchtime(tok.Str());
		lexer.yylex().Assert(Yytoken.EOL);
		break;
	      }
	    default:
	      lexer.Resync();
	      break;
	    }
	  }
	} catch (LexerException l) {
	  // java.lang.System.out.println("Parse Error - Continuing!");
	  lexer.Resync();
	}
      }
    } catch (java.io.IOException e) {
    }
  }
}

/**
 * Callback acts as a rendezvous for a particular command.
 */
class Callback {
  private int state;
  private String data;

  private SearchHandler searchhandler = null;

  private NotifyHandler notifyhandler = null;
  private String context = null;
  private boolean create;

  private Conn conn;

  public Callback(String context,
		  NotifyHandler notifyhandler,
		  boolean create) {
    this.context = context;
    this.notifyhandler = notifyhandler;
    this.create = create;
  }
  public Callback(SearchHandler searchhandler) {
    this.searchhandler = searchhandler;
  }
  public SearchHandler searchHandler() {
    return searchhandler;
  }
  public synchronized String Wait(Conn conn,
				  String tag,
				  Sendable cmd[])
    throws java.io.IOException, acapException {
      this.conn = conn;
      conn.sendCommand(cmd);
      conn.tags.put(tag, this);
      try {
	wait();
      } catch (java.lang.InterruptedException e) {
	// Not an error at all
      }
      conn.tags.remove(tag);
      if (state != Yytoken.OK)
	throw new acapException(data);
      return data;
  }
  public synchronized void Complete(int state, String data) {
    this.state = state;
    this.data = data;
    notify();
    // We do this here to ensure that the context handling will
    // be set up properly even if the next action requires the
    // new context state and no other threads get to run in the
    // meantime.
    if (context != null && state == Yytoken.OK) {
      if (create) {
	conn.contexts.put(context, notifyhandler);
      } else {
	conn.contexts.remove(context);
      }
    }
  }
}
